package smtpServer;

import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.net.Socket;
import java.util.Hashtable;
import java.util.logging.*;

/**
 * Deze klasse verzorgt de complete afhandeling met de cliënt.
 * Deze klasse kan gezien worden als de vervanger van de klasse SMTPProtocol.
 * 
 * Deze klasse werkt nouw samen met SMTPState, hier worden de commando's ontvangen
 * en vervolgens wordt via de SMTPState de juiste acties bij het commando gedaan.
 * 
 * @version 0.55
 * TODO [OPT] Implementeren van RFC1123, RFC1869 (deze voegen extra func. toe.). 
 */
public class SMTPClientHandler extends Thread {
	//Constanten
	//Verschillende acties (minimal implementation pag. 41)
	//Vergeet niet de methode buildCmdTable bij te werken!
	private final int CMD_HELO = 1001;
	private final int CMD_MAIL = 1002;
	private final int CMD_RCPT = 1003;
	private final int CMD_DATA = 1004;
	private final int CMD_RSET = 1005;
	private final int CMD_NOOP = 1006;
	private final int CMD_QUIT = 1007;
	//Commando om DATA Mode te beïndigen.
	private final int CMD_ENDDATA = 10008;
	private static Hashtable cmdTable;
	
	//Declaratie variabelen
	protected int clientID;
	protected Socket client;
	protected String hostname;
	protected BufferedReader input;
	protected PrintWriter output;
	protected SMTPState smtpState;
	protected boolean dataState;
	
	//Details van het bericht opslaan
	protected String msgFrom;
	protected String msgTo;
	
	//Logging
	private Logger log;

	/**
	 * Constructor
	 * @param client
	 * @param hostname
	 * @param clientID
	 */
	public SMTPClientHandler(Socket client, String hostname, int clientID){
		this.client = client;
		this.clientID = clientID;
		this.hostname = hostname;
		
		//Logger
		log = Logger.getLogger("SMTPServer");
		
		//Status goed zetten
		smtpState = new WAITState();
			
		//Commandotabel bijwerken
		buildCmdTable();
		
		//Writers aanmaken om te kunnen communiceren met de client.
		try{
			if (client != null){
				input = new BufferedReader(new InputStreamReader(client.getInputStream()));
				output = new PrintWriter(new OutputStreamWriter(client.getOutputStream()));
			}
		}
		catch(Exception ex){
			ex.printStackTrace();
			log.warning("client id: " + clientID + " Fout bij initialisatie.");
		}
			
	}
		
	/**
	 * Methode om de status te veranderen.
	 * @param smtpState
	 */
	public void setState(SMTPState smtpState){
		this.smtpState = smtpState;
	}
	
	/**
	 * De methode voor het afwerken van de cliënt
	 * De input wordt ontvangen, hier wordt het commando uitgehaald
	 * deze wordt vervolgens door het SMTPState object afgehandeld.
	 */
	public void run(){
		String cmdLine;
		String command;
		Integer cmd;
		int intCmd;
		String cmdArgs = null;
		String bfrOutput = "";

		log.info("client id: " + clientID + " Client thread gestart");
		
		if(input != null && output != null) {
			//Welkoms bericht
			output.println("220 "+hostname);
			output.flush();
			
			try {
				//Controleren of er iets ingevoerd is.
				while ((cmdLine = input.readLine()) != null){					
					//Initialisatie nieuwe commando
					command = cmdLine;
					bfrOutput = null;
					cmdArgs = null;
					
					//Als er een spatie in de input zit, dan commando eruit halen
					int intSpace = cmdLine.indexOf(" ");
					if (intSpace != -1){
						command = cmdLine.substring(0, intSpace);
						cmdArgs = cmdLine.substring(intSpace+1);
					}
					
					//Opzoeken commando in de tabel
					cmd = (Integer)cmdTable.get(command.toUpperCase());
					
					//Controleren of het commando bestaat.
					if (cmd == null){
						if (!smtpState.dataState){
							bfrOutput = "500 Ongeldige invoer.";
						}
					else if(smtpState.dataState){ 
							//addMsgData(cmdLine);
							smtpState.setMsgData(cmdLine);
						}
					}
					else{
						intCmd = cmd.intValue();
						switch(intCmd)
						{
							case CMD_HELO:{
								bfrOutput = smtpState.cmdHELO(this, cmdArgs, hostname);
								log.info("client id: " + clientID + " HELO Command.");
								break;
							}
							case CMD_MAIL: {
								bfrOutput = smtpState.cmdMAIL(this, cmdArgs);
								log.info("client id: " + clientID + " MAIL Command.");								
								break;
							}
							case CMD_RCPT: {
								bfrOutput = smtpState.cmdRCPT(this, cmdArgs);
								log.info("client id: " + clientID + " RCPT Command.");
								break;
							} 
							case CMD_DATA: {
								bfrOutput = smtpState.cmdDATA(this, cmdArgs);
								log.info("client id: " + clientID + " DATA Command.");
								break;								
							}
							case CMD_ENDDATA: {
								bfrOutput = smtpState.cmdEndData(this, cmdArgs);
								log.info("client id: " + clientID + " ENDDATA (.) Command.");
								break;
							}
							case CMD_RSET: {
								bfrOutput = smtpState.cmdRSET(this);
								log.info("client id: " + clientID + " RSET Command.");
								break;																
							} 
							case CMD_NOOP: {
								bfrOutput = smtpState.cmdNOOP();
								log.info("client id: " + clientID + " NOOP Command.");
								break;																								
							}
							case CMD_QUIT: {
								bfrOutput = smtpState.cmdQUIT();
								log.info("client id: " + clientID + " QUIT Command.");
								break;																								
							}
							
						}
						
					}
					if (bfrOutput != null){
						if (bfrOutput.equalsIgnoreCase("QUIT")){
							input.close();
							output.close();
							client.close();
							log.info("client id: " + clientID + " verbinding verbroken.");
							return;							
						} 
						else{
							output.println(bfrOutput);	
						}
						output.flush();
					}
				}
			}
			catch(Exception ex){
				ex.printStackTrace();
				log.warning("client id: " + clientID + " Fout in commando afhangeling opgetreden.");
			}

			
		}
	}	
	
	/**
	 * Deze methode bouwt een hashtable op met alle commands.
	 * Deze wordt gebruikt om ingevoerde commando's snel te detecteren.
	 */
	private void buildCmdTable(){
		cmdTable = new Hashtable();
		cmdTable.put("HELO", new Integer(CMD_HELO));	
		cmdTable.put("MAIL", new Integer(CMD_MAIL));
		cmdTable.put("RCPT", new Integer(CMD_RCPT));		
		cmdTable.put("DATA", new Integer(CMD_DATA));
		cmdTable.put("RSET", new Integer(CMD_RSET));	
		cmdTable.put("NOOP", new Integer(CMD_NOOP));
		cmdTable.put("QUIT", new Integer(CMD_QUIT));
		cmdTable.put(".", new Integer(CMD_ENDDATA));
	}
	
}
